#include <spu_intrinsics.h>
#include <spu_mfcio.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <filter.h>


// Block wird in Sub-Blcken dieser Gre abgearbeitet
#define BUFSIZE 2560

// Parameter-Kontext im lokalen Speicher
volatile parm_context ctx;

int main (unsigned long long spu_id, unsigned long long parm)
{
  int n;
  vector float *xp, *yp;
  vector unsigned char shp = {12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27};  // fr Shuffle

  // Filter-Koeffizienten im lokalen Speicher
  float a[2] __attribute__ ((aligned (16)));

  vector float a0v, a1v;  // Filter-Koeffizienten-Vektoren

  // Doppel-Puffer fr Quell- und Zieldaten
  float x0buf[BUFSIZE+4] __attribute__ ((aligned (16)));
  float x1buf[BUFSIZE+4] __attribute__ ((aligned (16)));
  float y0buf[BUFSIZE] __attribute__ ((aligned (16)));
  float y1buf[BUFSIZE] __attribute__ ((aligned (16)));
  // gerade Sub-Blocks verwenden Puffer x0,y0; ungerade verwenden x1,y1
  float *x = x0buf + 4, *y = y0buf, *x1 = x1buf + 4, *y1 = y1buf;
  // x lsst etwas Platz fr negative Indizes

  /* Tag-Ids fr alternierendes Double-Buffering.  0 fr gerade
     Sub-Blocks, 1 fr ungerade.  Werden nach jeder Iteration
     vertauscht. */
  unsigned int tag_id = 0, tag1_id = 1;

  int remaining;

  // Parameter-Kontext aus Hauptspeicher in lokalen Speicher laden:
  spu_writech(MFC_WrTagMask, -1);  // Maske = -1 heit: alle Tags
  //   Get-Kommando an MFC schicken
  //   Argumente: Hauptspeicher-Pointer, lokaler Pointer, Tag-Id, Kommando
  spu_mfcdma32((void *)(&ctx), (unsigned int)parm, sizeof(parm_context), tag_id, MFC_GET_CMD);
  (void)spu_mfcstat(MFC_TAG_UPDATE_ALL);  // Warten auf Beendigung der Get-Operation (Maske = alle)

  remaining = ctx.size;  // noch zu bearbeiten: size Floats

  // Quell-Daten fr ersten Sub-Block und Filter-Koeffizienten in lokalen Speicher laden:
  spu_writech(MFC_WrTagMask, 1 << tag_id);  // Maske = tag_id
  //   Get-Kommandos fr x und a an MFC schicken
  spu_mfcdma32((void *) (x-4), (unsigned int) (ctx.x-4), sizeof(float) * (BUFSIZE + 4), tag_id, MFC_GET_CMD);
  spu_mfcdma32((void *) a, (unsigned int) ctx.a, sizeof(float) * 2, tag_id, MFC_GET_CMD);
  (void)spu_mfcstat(MFC_TAG_UPDATE_ALL);  // Warten auf Beendigung (Maske = tag_id)

  a0v = spu_splats (a[0]);  // Koeffizienten-Vektoren initialisieren
  a1v = spu_splats (a[1]);

  while (remaining > 0)  // remaining wird um BUFSIZE verringert, bis alles abgearbeitet ist
  {
    int tmp;
    float *tmpp;

    // Sub-Block-Gre = Puffer-Gre, auer es sind nicht mehr so viele Floats abzuarbeiten
    int blksize = BUFSIZE;
    if (remaining < blksize)  blksize = remaining;

    // Wenn fr den nchsten Sub-Block noch etwas zu tun ist, dann:
    if (remaining > blksize)
    {
      // Gre des nchsten Sub-Blocks berechnen
      int nextblksize = BUFSIZE;
      if (remaining - blksize < nextblksize)  nextblksize = remaining - blksize;
      // MFC-Kommando zum Laden der Quell-Daten fr nchsten Sub-Block abschicken (andere Tag-Id!)
      spu_mfcdma32((void *) x1, (unsigned int) (ctx.x+blksize), sizeof(float) * nextblksize, tag1_id, MFC_GET_CMD);
      // Schwanz des aktuellen Sub-Blocks vor nchsten Subblock kopieren (fr negative Indizes)
      * (vector float *) (x1 - 4) = * (vector float *) (x + BUFSIZE - 4);
    }

    // SIMD-Algorithmus auf aktuellem Sub-Block
    xp = (vector float *) x;
    yp = (vector float *) y;
    for (n = 0; n < blksize/4; n ++)
    {
      *yp = *xp * a0v + spu_shuffle (xp[-1], *xp, shp) * a1v;
      yp ++;  xp ++;
    }

    // MFC-Kommando zum Sichern des Ziel-Puffers in den Hauptspeicher abschicken
    spu_mfcdma32((void *) y, (unsigned int) ctx.y, sizeof(float) * blksize, tag_id, MFC_PUT_CMD);

    /* Warten auf Beendigung der MFC-Kommandos fr die andere Tag-Id.
       Das heit, es wird darauf gewartet, dass das Laden der
       Quell-Daten fr den nchsten Sub-Block und das Sichern der
       Ziel-Daten vom letzten Sub-Block beendet ist. */
    spu_writech(MFC_WrTagMask, 1 << tag1_id);
    (void)spu_mfcstat(MFC_TAG_UPDATE_ALL);

    // Tag-Ids und Puffer-Pointer vertauschen zwecks Double-Buffering
    tmp = tag_id;  tag_id = tag1_id;  tag1_id = tmp;
    tmpp = x;  x = x1;  x1 = tmpp;
    tmpp = y;  y = y1;  y1 = tmpp;

    remaining -= blksize;  // blksize Floats wurden abgearbeitet
    ctx.x += blksize;  ctx.y += blksize;  // Hauptspeicher-Pointer um blksize Floats erhhen
  }

  // Warten auf Speichern des letzten Sub-Blocks
  spu_writech(MFC_WrTagMask, 1 << tag_id);
  (void)spu_mfcstat(MFC_TAG_UPDATE_ALL);

  return 0;
}
